from typing import Optional, List
from unittest import TestCase

from server.game.manager import ClientManager
from server.game.message import MessageToPlayer
from server.model.card import Value
from server.model.deck import Deck
from server.model.game import Game
from server.model.hand import Hand
from server.model.hands import pair, brelan
from server.model.known import CARRE_ACE
from server.model.players import Announce, Player


class MockPlayer(Player):

    def __init__(self,
                 name: str = None,
                 bets: List[Optional[Hand]] = None
                 ):
        super().__init__(name)

        if bets is None:
            bets = []
        self.bets: List[Optional[Hand]] = bets
        self.messages: List[MessageToPlayer] = []

    def count(self, msg: MessageToPlayer) -> int:
        return len([m for m in self.messages if m is msg])

    def announce(self, current_bet: Optional[Hand]) -> Announce:
        if self.bets:
            return Announce(self.bets.pop())
        else:
            return Announce(CARRE_ACE)

    def receive(self, message: MessageToPlayer):
        self.messages.append(message)

    def print_msgs(self) -> str:
        '|'.join([str(m) for m in self.messages])


class MockManager(ClientManager):
    def __init__(self, players: List[MockPlayer]):
        super().__init__()
        self.players = players

    def send(self, to: Player, message: MessageToPlayer, extra=None):
        if isinstance(to, MockPlayer):
            to.receive(message)
            print(f"Sent {message} to {to}.")


class TestManager(TestCase):

    def setUp(self) -> None:
        self.j1 = MockPlayer("j1", [pair(Value.Ace)])
        self.j2 = MockPlayer("j2", [brelan(Value.Two)])
        self.manager = MockManager([self.j1, self.j2])
        cards = []
        cards.extend(self.j1.bets[0].cards)
        cards.extend(self.j2.bets[0].cards)
        self.game = Game(players=[self.j1, self.j2],
                         deck=Deck(cards=cards),
                         manager=self.manager)

    def test_turn_messages(self):
        self.game.new_turn()
        self.assertEqual(self.j1.count(MessageToPlayer.LoseRound), 6, f"j1 should lose 6 rounds.")
        self.assertEqual(self.j2.count(MessageToPlayer.LoseRound), 6, f"j2 should lose 6 rounds.")

        for player in [self.j1, self.j2]:
            self.assertIn(MessageToPlayer.Lose, player.messages, "Loser not announced")

    def test_game_messages(self):
        self.game.new_game()
        self.assertEqual(len([m for m in self.j1.messages if m is MessageToPlayer.LoseRound]), 5,
                         f"{self.j1} should lose 5 rounds: {'|'.join([str(m) for m in self.j1.messages])}")

        self.assertEqual(self.j1.count(MessageToPlayer.YourTurn), 6,
                         f"{self.j1} should play 6 rounds: {self.j1.print_msgs()}")

        self.assertEqual(self.j2.count(MessageToPlayer.YourTurn), 1,
                         f"{self.j2} should play 1 rounds: {self.j2.print_msgs()}")

        self.assertEqual(len([m for m in self.j2.messages if m is MessageToPlayer.LoseRound]), 0,
                         f"{self.j2} should lose 0 rounds: {'|'.join([str(m) for m in self.j2.messages])}")
        for player in [self.j1, self.j2]:
            self.assertEqual(player.count(MessageToPlayer.Announce), 7,
                             f"{player} should see 7 announces: {player.print_msgs()}")
        self.assertIn(MessageToPlayer.WinnerIs, player.messages, "Winner not announced")
        self.assertIn(MessageToPlayer.Win, player.messages, "Win not told")
